/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*               This file is part of the program and library                */
/*    PaPILO --- Parallel Presolve for Integer and Linear Optimization       */
/*                                                                           */
/* Copyright (C) 2020-2025 Zuse Institute Berlin (ZIB)                       */
/*                                                                           */
/* Licensed under the Apache License, Version 2.0 (the "License");           */
/* you may not use this file except in compliance with the License.          */
/* You may obtain a copy of the License at                                   */
/*                                                                           */
/*     http://www.apache.org/licenses/LICENSE-2.0                            */
/*                                                                           */
/* Unless required by applicable law or agreed to in writing, software       */
/* distributed under the License is distributed on an "AS IS" BASIS,         */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  */
/* See the License for the specific language governing permissions and       */
/* limitations under the License.                                            */
/*                                                                           */
/* You should have received a copy of the Apache-2.0 license                 */
/* along with PaPILO; see the file LICENSE. If not visit scipopt.org.        */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef _PAPILO_CMAKE_CONFIG_HPP_
#define _PAPILO_CMAKE_CONFIG_HPP_

#define PAPILO_HAVE_FLOAT128
#define PAPILO_HAVE_GMP
#define PAPILO_COMMAND_LINE_AVAILABLE
#define PAPILO_SERIALIZATION_AVAILABLE
/* #undef PAPILO_HAVE_LUSOL */
#define PAPILO_USE_STANDARD_HASHMAP
#define PAPILO_USE_BOOST_IOSTREAMS_WITH_ZLIB
#define PAPILO_USE_BOOST_IOSTREAMS_WITH_BZIP2
#define PAPILO_GITHASH_AVAILABLE
#define PAPILO_TBB

#define PAPILO_VERSION_MAJOR 3
#define PAPILO_VERSION_MINOR 0
#define PAPILO_VERSION_PATCH 0
#define PAPILO_VERSION_TWEAK 0
#define PAPILO_API_VERSION 11

#ifdef PAPILO_HAVE_GMP
   #define GMP_VERSION "6.3.0"
#endif

#ifdef PAPILO_GITHASH_AVAILABLE
   #define PAPILO_GITHASH "4cbdc327"
#endif

#endif
