/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*               This file is part of the program and library                */
/*    PaPILO --- Parallel Presolve for Integer and Linear Optimization       */
/*                                                                           */
/* Copyright (C) 2020-2025 Zuse Institute Berlin (ZIB)                       */
/*                                                                           */
/* Licensed under the Apache License, Version 2.0 (the "License");           */
/* you may not use this file except in compliance with the License.          */
/* You may obtain a copy of the License at                                   */
/*                                                                           */
/*     http://www.apache.org/licenses/LICENSE-2.0                            */
/*                                                                           */
/* Unless required by applicable law or agreed to in writing, software       */
/* distributed under the License is distributed on an "AS IS" BASIS,         */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  */
/* See the License for the specific language governing permissions and       */
/* limitations under the License.                                            */
/*                                                                           */
/* You should have received a copy of the Apache-2.0 license                 */
/* along with PaPILO; see the file LICENSE. If not visit scipopt.org.        */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef _PAPILO_MISC_FLAGS_HPP_
#define _PAPILO_MISC_FLAGS_HPP_

#include <cstdint>
#include <type_traits>

namespace papilo
{

template <typename BaseType>
class Flags
{
 public:
   Flags( BaseType t ) : state( static_cast<UnderlyingType>( t ) ) {}

   Flags() : state( 0 ) {}

   template <typename... Args>
   void
   set( Args... flags )
   {
      state |= joinFlags( flags... );
   }

   void
   clear()
   {
      state = 0;
   }

   template <typename... Args>
   void
   unset( Args... flags )
   {
      state &= ~joinFlags( flags... );
   }

   template <typename... Args>
   bool
   test( Args... flags ) const
   {
      return state & joinFlags( flags... );
   }

   template <typename... Args>
   bool
   equal( Args... flags ) const
   {
      return state == joinFlags( flags... );
   }

   template <typename Archive>
   void
   serialize( Archive& ar, const unsigned int version )
   {
      ar& state;
   }

 private:
   using UnderlyingType = typename std::underlying_type<BaseType>::type;

   static UnderlyingType
   joinFlags( BaseType f1 )
   {
      return static_cast<UnderlyingType>( f1 );
   }

   template <typename... Args>
   static UnderlyingType
   joinFlags( BaseType f1, Args... other )
   {
      return static_cast<UnderlyingType>( f1 ) | joinFlags( other... );
   }

   UnderlyingType state;
};

} // namespace papilo

#endif
